__version__ = "0.1.4"

from .application import Pykour
from .config import Config
from .request import Request
from .response import Response
from .router import Router
from .url import URL

__all__ = ["__version__", "Pykour", "Router", "Request", "Response", "URL", "Config"]



import json
from typing import Mapping, Any, Iterator, cast, List, Tuple, Union
from collections import defaultdict

from pykour.types import Scope, Receive
from pykour.url import URL


class Request(Mapping[str, Any]):
    """Request is a class that represents a request from a client."""

    def __init__(self, scope: Scope, receive: Receive):
        """Initializes a new instance of the Request class.

        Args:
            scope: The ASGI scope.
            receive: The ASGI receive function.
        """

        self.scope = scope
        self.receive = receive
        self._headers = defaultdict(list)
        self.content_type = None
        self.charset = "utf-8"
        self.path_params: dict[str, str] = {}

        if "headers" in self.scope:
            for key, value in self.scope["headers"]:
                decoded_key = key.decode("latin1").lower()
                decoded_value = value.decode("latin1")
                self._headers[decoded_key].append(decoded_value)

                if decoded_key == "content-type":
                    self.content_type = decoded_value
                    if "charset=" in decoded_value:
                        self.charset = decoded_value.split("charset=")[-1]

        self._stream_consumed = False

    def __getitem__(self, key: str) -> Any:
        return self.scope[key]

    def __iter__(self) -> Iterator[Any]:
        return iter(self.scope)

    def __len__(self) -> int:
        return len(self.scope)

    __eq__ = object.__eq__
    __hash__ = object.__hash__

    @property
    def app(self) -> Any:
        """Returns the ASGI application instance.

        Returns:
            Application name.
        """
        return self.scope["app"]

    @property
    def url(self) -> URL:
        """Returns the URL instance.

        Returns:
            URL instance.
        """
        return URL(scope=self.scope)

    @property
    def headers(self) -> dict[str, list[str]]:
        """Returns the headers.

        Returns:
            Headers.
        """
        return self._headers

    def get_header(self, name: str) -> list[str]:
        """Returns the header value.

        Args:
            name: The header name.
        """
        return self._headers.get(name)

    @property
    def method(self) -> Union[str, None]:
        """Returns the HTTP method.

        Returns:
            HTTP method.
        """
        if "method" not in self.scope:
            return None
        else:
            return cast(str, self.scope["method"])

    @property
    def scheme(self) -> Union[str, None]:
        """Returns the URL scheme."""

        if "scheme" not in self.scope:
            return None
        else:
            return cast(str, self.scope["scheme"]).upper()

    @property
    def version(self) -> Union[str, None]:
        """Returns the HTTP version.

        Returns:
            HTTP version.
        """
        if "http_version" not in self.scope:
            return None
        else:
            return self.scope["http_version"]

    @property
    def client(self) -> Union[str, None]:
        """Returns the client address.

        Returns:
            Client address.
        """
        if "client" not in self.scope:
            return None
        else:
            client = self.scope["client"]
            return f"{client[0]}:{client[1]}"

    @property
    def path(self) -> str:
        """Returns the request path.

        Returns:
            Request path.
        """
        return self.scope["path"]

    @property
    def query_string(self) -> str:
        """Returns the query string.

        Returns:
            Query string.
        """
        return self.scope["query_string"]

    @property
    def accept(self) -> List[str]:
        """Returns the accept header.

        Returns:
            Accept header.
        """
        return self.get_sorted_accept_list(self.scope)

    async def body(self) -> bytes:
        """Reads the request body.

        Returns:
            The request body.
        """
        try:
            body = b""
            more_body = True

            while more_body:
                message = await self.receive()
                body += message.get("body", b"")
                more_body = message.get("more_body", False)

            return body
        except Exception as e:
            print(f"Error occurred while receiving body: {e}")
            raise e

    async def json(self) -> Any:
        """Parses the request body as JSON.

        Returns:
            The parsed JSON object.
        """
        try:
            body = await self.body()
            return json.loads(body)
        except Exception as e:
            print(f"Error occurred while parsing JSON: {e}")
            raise e

    @staticmethod
    def parse_accept_header(accept_header: str) -> List[Tuple[str, float]]:
        """
        Parse the accept header and return a sorted list of MIME types.
        """
        accept_values = accept_header.split(",")
        result = []

        for value in accept_values:
            parts = value.split(";")
            mime_type = parts[0].strip()
            q_value = 1.0

            if len(parts) > 1 and parts[1].strip().startswith("q="):
                try:
                    q_value = float(parts[1].strip()[2:])
                except ValueError:
                    pass

            result.append((mime_type, q_value))

        result.sort(key=lambda x: x[1], reverse=True)
        return result

    def get_sorted_accept_list(self, scope: Scope) -> List[str]:
        """
        Parse the accept header and return a sorted list of MIME types.
        """
        headers = scope.get("headers", [])
        accept_header = ""

        for header in headers:
            if header[0].decode("latin-1") == "accept":
                accept_header = header[1].decode("latin-1")
                break

        if not accept_header:
            return []

        parsed_accept = self.parse_accept_header(accept_header)
        return [mime_type for mime_type, q_value in parsed_accept]



import importlib
from typing import Optional, Dict, Any, List, Union

from pykour.config import Config


class Connection:
    def __init__(self, db_type, **kwargs):
        self.db_type = db_type
        self.conn = None
        if self.db_type == "sqlite":
            sqlite3 = importlib.import_module("sqlite3")
            self.conn = sqlite3.connect(kwargs["url"])
        else:
            raise ValueError(f"Unsupported session type: {self.db_type}")

        self.cursor = self.conn.cursor()

    @classmethod
    def from_config(cls, config: Config):
        db_type = config.get_datasource_type()
        url = config.get_datasource_url()
        username = config.get_datasource_username()
        password = config.get_datasource_password()
        return cls(db_type, url=url, username=username, password=password)

    def find(self, query: str, params: Optional[Dict[str, Any]] = None) -> Union[Dict[str, Any], None]:
        self._execute(query, params)
        row = self.cursor.fetchone()
        if row:
            columns = [desc[0] for desc in self.cursor.description]
            return dict(zip(columns, row))
        return None

    def select(self, query: str, params: Optional[Dict[str, Any]] = None) -> List[Dict[str, Any]]:
        self._execute(query, params)
        rows = self.cursor.fetchall()
        columns = [desc[0] for desc in self.cursor.description]
        return [dict(zip(columns, row)) for row in rows]

    def execute(self, query: str, params: Optional[Dict[str, Any]] = None) -> int:
        self._execute(query, params)
        return self.cursor.rowcount

    def commit(self):
        self.conn.commit()

    def rollback(self):
        self.conn.rollback()

    def close(self):
        if self.cursor:
            self.cursor.close()
            self.cursor = None
        if self.conn:
            self.conn.close()
            self.conn = None

    def _execute(self, query, params=None):
        if params:
            self.cursor.execute(query, params)
        else:
            self.cursor.execute(query)



from __future__ import annotations
from abc import ABCMeta
from typing import Any, Dict, get_type_hints, Type, get_origin, get_args


class BaseSchemaMetaclass(ABCMeta):
    def __new__(mcs, cls_name: str, bases, attrs, **kwargs: Any) -> type:
        fields = {key: value for key, value in attrs.items() if not key.startswith("_")}
        validators = {key: value for key, value in attrs.items() if hasattr(value, "__validator__")}

        cls = super().__new__(mcs, cls_name, bases, attrs, **kwargs)
        cls.__fields__ = fields  # type: ignore[attr-defined]
        cls.__validators__ = validators  # type: ignore[attr-defined]
        return cls


class BaseSchema(metaclass=BaseSchemaMetaclass):
    """Base class for schema."""

    def __init__(self, /, **data: Any) -> None:
        self._validate(data)
        self._set_attributes(data)

    def _set_attributes(self, data: Dict[str, Any]) -> None:
        for key, value in data.items():
            setattr(self, key, value)

    def _validate(self, data: Dict[str, Any]) -> None:
        for validator_name, validator_fn in self.__class__.__validators__.items():  # type: ignore[attr-defined]
            field_name = validator_fn.__validator__
            if field_name in data:
                value = data[field_name]
                validator_fn(self, value)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> BaseSchema:
        """
        Create instance from dictionary.

        Args:
            data: Dictionary with data.
        Returns:
            Instance of schema.
        """
        init_data = {}
        field_types = get_type_hints(cls)

        for field, field_type in field_types.items():
            if field in data:
                value = data[field]
                init_data[field] = cls._convert_value(field_type, value)
            else:
                init_data[field] = None

        return cls(**init_data)

    @staticmethod
    def _convert_value(field_type: Type, value: Any) -> Any:
        origin = get_origin(field_type)
        args = get_args(field_type)

        if isinstance(value, dict):
            if origin is not None and issubclass(origin, dict):
                return value
            elif isinstance(field_type, type) and issubclass(field_type, BaseSchema):
                return field_type.from_dict(value)
            else:
                return value
        elif isinstance(value, list) and origin is list:
            inner_type = args[0]
            return [BaseSchema._convert_value(inner_type, item) for item in value]
        else:
            return value



import argparse

from gunicorn.app.base import BaseApplication
import uvicorn
import signal
import sys
import asyncio
from functools import partial
import os

from pykour import __version__

usage_text = """usage: pykour [-h] {subcommand} ...

positional arguments:
    run                 Run Web Server

optional arguments:
  -h, --help            show this help message and exit
"""

sys.path.append(os.getcwd())


class PykourArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        """Ignore unknown arguments."""
        ...


def signal_handler(_signal, _frame):
    sys.exit(0)


def parse_args(args):
    parser = PykourArgumentParser(description="Pykour CLI", add_help=False)

    parser.add_argument("-v", "--version", action="version", version=f"Pykour v{__version__}")
    parser.add_argument("-h", "--help", action="store_true", help="show this help message and exit")

    subparsers = parser.add_subparsers(dest="command")

    # Add the 'dev' subcommand
    dev_parser = subparsers.add_parser("dev", help="Run Web Server for Development")
    dev_parser.add_argument("app", type=str, help="The ASGI app instance to run, e.g., main:app")
    dev_parser.add_argument("--host", type=str, default="127.0.0.1", help="The host to bind to")
    dev_parser.add_argument("--port", type=int, default=8000, help="The port to bind to")
    dev_parser.add_argument("--reload", action="store_true", default=True, help="Enable auto-reload")
    dev_parser.add_argument("--workers", type=int, default=1, help="Number of worker processes")

    # Add the 'run' subcommand
    run_parser = subparsers.add_parser("run", help="Run Web Server for Production")
    run_parser.add_argument("app", type=str, help="The ASGI app instance to run, e.g., main:app")
    run_parser.add_argument("--host", type=str, default="0.0.0.0", help="The host to bind to")
    run_parser.add_argument("--port", type=int, default=8000, help="The port to bind to")
    run_parser.add_argument("--reload", action="store_true", default=False, help="Enable auto-reload")
    run_parser.add_argument("--workers", type=int, default=(os.cpu_count() * 2) + 1, help="Number of worker processes")

    # Parse the arguments
    return parser.parse_args(args)


def main(args=None):
    loop = asyncio.get_event_loop()
    for sig in (signal.SIGINT, signal.SIGTERM):
        loop.add_signal_handler(sig, partial(signal_handler, sig, None))

    if args is None:
        args = sys.argv[1:]
    args = parse_args(args)

    if args.help:
        print(usage_text)
        sys.exit(0)
    elif args.command == "dev":
        os.environ["PYKOUR_ENV"] = "development"

        uvicorn.run(
            args.app,
            host=args.host,
            port=args.port,
            reload=args.reload,
            workers=args.workers,
            server_header=False,
        )
    elif args.command == "run":
        os.environ["PYKOUR_ENV"] = "production"

        class StandaloneApplication(BaseApplication):
            def __init__(self, app, opts=None):
                self.app = app
                self.options = opts or {}
                super().__init__()

            def load_config(self):
                config = {
                    key: value for key, value in self.options.items() if key in self.cfg.settings and value is not None
                }
                for key, value in config.items():
                    self.cfg.set(key.lower(), value)

            def load(self):
                return self.app

        options = {
            "bind": args.host + ":" + str(args.port),
            "workers": args.workers,
            "worker_class": "uvicorn_worker.UvicornWorker",
            "reload": args.reload,
        }

        StandaloneApplication(
            args.app,
            options,
        ).run()
    else:
        print(usage_text)
        sys.exit(1)


if __name__ == "__main__":
    main()



from queue import Queue, Empty, Full
from threading import Lock

from pykour.config import Config
from pykour.db.connection import Connection


class ConnectionPool:
    def __init__(self, config: Config):
        self.config = config
        self.max_connections = config.get_datasource_pool_max_connections()
        self.pool: Queue = Queue(maxsize=self.max_connections)
        self.lock = Lock()
        for _ in range(self.max_connections):
            self.pool.put(self._create_new_connection())

    def get_connection(self):
        try:
            return self.pool.get_nowait()
        except Empty:
            with self.lock:
                return self._create_new_connection()

    def release_connection(self, connection):
        try:
            self.pool.put_nowait(connection)
        except Full:
            connection.close()

    def close_all_connections(self):
        while not self.pool.empty():
            conn = self.pool.get_nowait()
            conn.close()

    def _create_new_connection(self):
        return Connection.from_config(self.config)

    def __del__(self):
        self.close_all_connections()



from pykour.types import ASGIApp, Scope, Receive, Send


class BaseMiddleware:
    """Base middleware class."""

    def __init__(self, app: ASGIApp):
        """
        Initialize the middleware with the application.

        Args:
            app: The ASGI application.
        """
        self.app = app

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ):
        """
        Call the middleware.

        Args:
            scope: The ASGI scope.
            receive: The ASGI receive channel.
            send: The ASGI send channel.
        """

        await self.process_request(scope, receive, send)
        await self.app(scope, receive, send)
        await self.process_response(scope, receive, send)

    async def process_request(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ):
        """
        Executed before the request is processed.

        Args:
            scope: The ASGI scope.
            receive: The ASGI receive channel.
            send: The ASGI send channel.
        """
        ...

    async def process_response(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ):
        """
        Executed after response processing.

        Args:
            scope: The ASGI scope.
            receive: The ASGI receive channel.
            send: The ASGI send channel.
        """
        ...



from http import HTTPStatus

from pykour.types import Send, HTTPStatusCode


class Response:
    """Response is a class that represents an HTTP response."""

    def __init__(
        self,
        send: Send,
        status_code: HTTPStatusCode = HTTPStatus.OK,
        charset: str = "utf-8",
        content_type: str = "application/json",
    ) -> None:
        """Create a new Response object.

        Args:
            send: The send function of the ASGI application.
            status_code: The status code of the response.
            charset: The charset of the response.
            content_type: The content type of the response.
        """

        self.send = send
        self._status_code = status_code
        self._charset = charset
        self._content_type = content_type
        self._headers = []
        self._headers.append(("Content-Type", f"{content_type}; charset={charset}"))
        self._content = ""

    @property
    def status(self) -> HTTPStatusCode:
        """Get the status code of the response.

        Returns:
            The status code of the response.
        """
        return self._status_code

    @status.setter
    def status(self, status_code: HTTPStatusCode) -> None:
        """Set the status code of the response.

        Args:
            status_code: The status code of the response.
        """

        self._status_code = status_code

    @property
    def charset(self) -> str:
        """Get the charset of the response.

        Returns:
            The charset of the response.
        """
        return self._charset

    @charset.setter
    def charset(self, charset: str) -> None:
        """Set the charset of the response.

        Args:
            charset: The charset of the response.
        """
        self._charset = charset
        self._headers[0] = ("Content-Type", f"{self._content_type}; charset={charset}")

    @property
    def content_type(self) -> str:
        """Get the content type of the response.

        Returns:
            The content type of the response.
        """

        return self._content_type

    @content_type.setter
    def content_type(self, content_type: str) -> None:
        """Set the content type of the response.

        Args:
            content_type: The content type of the response.
        """

        self._content_type = content_type
        self._headers[0] = ("Content-Type", f"{content_type}; charset={self._charset}")

    @property
    def headers(self) -> list[tuple[str, str]]:
        """Get the headers of the response.

        Returns:
            The headers of the response.
        """
        return self._headers

    def get_header(self, key: str) -> list[str]:
        """Get the value of a header.

        Args:
            key: The key of the header.
        """

        result = []
        for header in self._headers:
            if header[0] == key:
                result.append(header[1])
        return result

    def add_header(self, key: str, value: str) -> None:
        """Add a header to the response.

        Args:
            key: The key of the header.
            value: The value of the header.
        """
        self._headers.append((key, value))

    @property
    def content(self) -> str:
        """Get the content of the response.

        Returns:
            The content of the response.
        """
        return self._content

    @content.setter
    def content(self, content: str) -> None:
        """Set the content of the response.

        Args:
            content: The content of the response.
        """
        self._content = content

    async def render(self) -> None:
        """Render the response."""
        await self.send(
            {
                "type": "http.response.start",
                "status": self._status_code,
                "headers": self._headers,
            }
        )
        if self._status_code == HTTPStatus.NO_CONTENT or self.content is None or self.content == "":
            await self.send({"type": "http.response.body", "body": b""})
        else:
            await self.send({"type": "http.response.body", "body": self._content.encode(self._charset)})



from __future__ import annotations

from typing import Any
from urllib.parse import (
    SplitResult,
    parse_qsl,
    urlsplit,
)

from pykour.types import Scope


class URL:
    """URL is a class that represents a URL."""

    def __init__(self, url: str = None, scope: Scope = None) -> None:
        """URL is a class that represents a URL.

        Args:
            url: The URL.
            scope: The ASGI scope.
        """

        if not url and not scope:
            raise ValueError("Either 'url' or 'scope' must be provided.")

        if not url and scope:
            scheme = scope.get("scheme", "http")
            server = scope.get("server", None)
            path = scope["path"]
            query_string = scope.get("query_string", b"")

            host_header = None
            for key, value in scope["headers"]:
                if key == b"host":
                    host_header = value.decode("latin-1")
                    break

            if host_header is not None:
                url = f"{scheme}://{host_header}{path}"
            elif server:
                host, port = server
                default_port = {"http": 80, "https": 443}[scheme]
                if port == default_port:
                    url = f"{scheme}://{host}{path}"
                else:
                    url = f"{scheme}://{host}:{port}{path}"
            else:
                raise ValueError("Could not determine the URL.")

            if query_string:
                url += "?" + query_string.decode()

        self.url = url

    @property
    def components(self) -> SplitResult:
        """Returns the URL components.

        Returns:
            The URL components.
        """
        return urlsplit(self.url)

    @property
    def scheme(self) -> str:
        """Returns the scheme.

        Returns:
            The scheme.
        """
        return self.components.scheme

    @property
    def hostname(self) -> str:
        """Returns the hostname.

        Returns:
            The hostname.
        """
        return self.components.hostname

    @property
    def port(self) -> int:
        """Returns the port number.

        Returns:
            The port number.
        """
        return self.components.port or {"http": 80, "https": 443}[self.scheme]

    @property
    def path(self) -> str:
        """Returns the path.

        Returns:
            The path.
        """
        return self.components.path or "/"

    @property
    def query(self) -> str:
        """Returns the query string.

        Returns:
            The query string.
        """
        return self.components.query

    @property
    def is_secure(self) -> bool:
        """Returns True if the URL is secure.

        Returns:
            True if the URL is secure.
        """
        return self.scheme in ("https", "wss")

    @property
    def query_params(self) -> dict:
        return dict(parse_qsl(self.query))

    def __eq__(self, other: Any) -> bool:
        return str(self) == str(other)

    def __str__(self) -> str:
        return self.url

    def __repr__(self) -> str:
        url = str(self)
        return f"{self.__class__.__name__}({repr(url)})"



from __future__ import annotations
from http import HTTPStatus
from typing import Any, Callable, Dict, Union, List


class Route:
    def __init__(self, path: str, method: str, handler: Any):
        self.path = path
        self.method = method
        self.handler = handler
        self.path_params: Dict[str, str] = {}

    def set_path_params(self, path_params: Dict[str, str]):
        self.path_params = path_params


class Node:
    def __init__(self, part: str, is_wild: bool = False):
        self.part = part
        self.children: list[Node] = []
        self.is_wild = is_wild
        self.route: Union[Route, None] = None

    def insert(self, pattern: str, route: Route):
        parts = pattern.strip("/").split("/")

        node = self
        for part in parts:
            child = node.match_child(part)
            if not child:
                child = Node(
                    part,
                    part.startswith(":") or part.startswith("*") or part.startswith("{"),
                )
                node.children.append(child)
            node = child

        node.route = route

    def search(self, path: str):
        parts = path.strip("/").split("/")
        path_params = {}

        node = self
        for part in parts:
            child = node.match_child(part)
            if not child:
                return None, {}
            if child.is_wild:
                var_name = child.part.lstrip(":*{").rstrip("}")
                path_params[var_name] = part
            node = child

        # route = node.route_map.get(method)
        return node.route, path_params

    def match_child(self, part: str):
        for child in self.children:
            if child.part == part and not child.is_wild:
                return child
        for child in self.children:
            if child.is_wild:
                return child
        return None


class Router:
    SUPPORTED_METHODS = ["GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS", "HEAD"]

    def __init__(self, prefix: str = ""):
        """Router class.

        Args:
            prefix: Prefix for the router.
        """
        self.roots: dict[str, Node] = {}
        self.prefix = prefix.rstrip("/")

    def __str__(self):
        routes: List[str] = []

        def traverse(node, prefix=""):
            for child in node.children:
                new_prefix = f"{prefix}/{child.part}".strip("/")
                route = child.route
                if route:
                    handler_name = route.handler[0].__name__
                    route_description = f"{method} /{new_prefix} -> {handler_name}()"
                    routes.append(route_description)
                traverse(child, new_prefix)

        for method, root in self.roots.items():
            traverse(root)
        return "\n".join(routes)

    def __repr__(self):
        return "Router(prefix='{}')".format(self.prefix)

    def get(self, path: str, status_code: HTTPStatus = HTTPStatus.OK) -> Callable:
        """Decorator for GET method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="GET", status_code=status_code)

    def post(self, path: str, status_code: HTTPStatus = HTTPStatus.CREATED) -> Callable:
        """Decorator for POST method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="POST", status_code=status_code)

    def put(self, path: str, status_code: HTTPStatus = HTTPStatus.OK) -> Callable:
        """Decorator for PUT method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="PUT", status_code=status_code)

    def delete(self, path: str, status_code: HTTPStatus = HTTPStatus.NO_CONTENT) -> Callable:
        """Decorator for DELETE method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="DELETE", status_code=status_code)

    def patch(self, path: str, status_code: HTTPStatus = HTTPStatus.OK) -> Callable:
        """Decorator for PATCH method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="PATCH", status_code=status_code)

    def options(self, path: str, status_code: HTTPStatus = HTTPStatus.OK) -> Callable:
        """Decorator for OPTIONS method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="OPTIONS", status_code=status_code)

    def head(self, path: str, status_code: HTTPStatus = HTTPStatus.OK) -> Callable:
        """Decorator for HEAD method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="HEAD", status_code=status_code)

    def route(self, path: str, method: str = "GET", status_code: Union[HTTPStatus, int] = HTTPStatus.OK) -> Callable:
        """Decorator for route.

        Args:
            path: URL path.
            method: HTTP method.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """

        if method not in self.SUPPORTED_METHODS:
            raise ValueError(f"Unsupported HTTP method: {method}")

        def decorator(func):
            self.add_route(path, method, (func, status_code))
            return func

        return decorator

    def add_router(self, router: Router, prefix: str = ""):
        """Add router.

        Args:
            router: Router instance.
            prefix: Prefix for the router.
        """

        def traverse(node, prefix=""):
            for child in node.children:
                new_prefix = f"{prefix}/{child.part}".strip("/")
                route = child.route
                if route:
                    self.add_route(new_prefix, method, route.handler)
                traverse(child, new_prefix)

        for method, root in router.roots.items():
            traverse(root, prefix=prefix)

    def add_route(self, path: str, method: str, handler: Any):
        """Add route.

        Args:
            path: URL path.
            method: HTTP method.
            handler: Route handler.
        """
        if self.prefix:
            full_path = f"/{self.prefix}{path}"
        else:
            full_path = path
        route = Route(full_path, method, handler)
        if method not in self.roots:
            self.roots[method] = Node("")
        self.roots[method].insert(full_path, route)

    def get_route(self, path: str, method: str) -> Union[Route, None]:
        """Get route.

        Args:
            path: URL path.
            method: HTTP method.
        """
        if method in self.roots:
            route, path_params = self.roots[method].search(path)
            if route:
                route.set_path_params(path_params)
            return route
        return None

    def get_allowed_methods(self, path: str) -> List[str]:
        """Get allowed HTTP methods for the specified path.

        Args:
            path: URL path.

        Returns:
            List of allowed HTTP methods.
        """
        allowed_methods = []
        http_methods = [
            "GET",
            "POST",
            "PUT",
            "DELETE",
            "PATCH",
            "OPTIONS",
            "HEAD",
        ]

        for method in http_methods:
            if self.get_route(path, method):
                allowed_methods.append(method)

        return allowed_methods

    def exists(self, path: str, method: str) -> bool:
        """Check if route exists.

        Args:
            path: URL path.
            method: HTTP method.
        Returns:
            True if route exists, False otherwise.
        """
        return self.get_route(path, method) is not None



import logging
import time
import json
import os
from http import HTTPStatus
from typing import Callable, Any

import pykour.exceptions as ex
from pykour.config import Config
from pykour.call import call
from pykour.db.pool import ConnectionPool
from pykour.logging import setup_logging, write_access_log


from pykour.request import Request
from pykour.response import Response
from pykour.router import Router
from pykour.types import Scope, Receive, Send, ASGIApp, HTTPStatusCode
from colorama import Fore


STATUS_COLORS = {
    "2xx": Fore.GREEN,
    "3xx": Fore.BLUE,
    "4xx": Fore.YELLOW,
    "5xx": Fore.RED,
}


class Pykour:
    SUPPORTED_METHODS = ["GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS", "HEAD"]

    def __init__(self, config: str = None):
        self._config = Config(config) if config else Config()
        setup_logging(self._config.get_log_levels())
        self.production_mode = os.getenv("PYKOUR_ENV") == "production"
        self.router = Router()
        self.app: ASGIApp = RootASGIApp()
        self.logger = logging.getLogger("pykour")
        if self._config.get_datasource_type():
            self.pool = ConnectionPool(self._config)
        else:
            self.pool = None

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["app"] = self
        await self.app(scope, receive, send)

    @property
    def config(self) -> Config:
        return self._config

    def add_middleware(self, middleware: Callable, **kwargs: Any) -> None:
        """Add middleware to the application.

        Args:
            middleware: Middleware class.
            kwargs: Middleware arguments.
        """

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(f"Add middleware: {middleware.__name__}")
        self.app = middleware(self.app, **kwargs)

    def get(self, path: str, status_code: HTTPStatusCode = HTTPStatus.OK) -> Callable:
        """Decorator for GET method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="GET", status_code=status_code)

    def post(self, path: str, status_code: HTTPStatusCode = HTTPStatus.CREATED) -> Callable:
        """Decorator for POST method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="POST", status_code=status_code)

    def put(self, path: str, status_code: HTTPStatusCode = HTTPStatus.OK) -> Callable:
        """Decorator for PUT method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="PUT", status_code=status_code)

    def delete(self, path: str, status_code: HTTPStatusCode = HTTPStatus.NO_CONTENT) -> Callable:
        """Decorator for DELETE method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="DELETE", status_code=status_code)

    def patch(self, path: str, status_code: HTTPStatusCode = HTTPStatus.OK) -> Callable:
        """Decorator for PATCH method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="PATCH", status_code=status_code)

    def options(self, path: str, status_code: HTTPStatusCode = HTTPStatus.OK) -> Callable:
        """Decorator for OPTIONS method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="OPTIONS", status_code=status_code)

    def head(self, path: str, status_code: HTTPStatusCode = HTTPStatus.OK) -> Callable:
        """Decorator for HEAD method.

        Args:
            path: URL path.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """
        return self.route(path=path, method="HEAD", status_code=status_code)

    def route(self, path: str, method: str = "GET", status_code: HTTPStatusCode = HTTPStatus.OK) -> Callable:
        """Decorator for route.

        Args:
            path: URL path.
            method: HTTP method.
            status_code: HTTP status code.
        Returns:
            Route decorator.
        """

        if method not in self.SUPPORTED_METHODS:
            raise ValueError(f"Unsupported HTTP Method: {method}")

        def decorator(func):
            self.router.add_route(path=path, method=method, handler=(func, status_code))
            return func

        return decorator

    def add_router(self, router: Router, prefix: str = "") -> None:
        """Add a router to the application.

        Args:
            router: Router object.
            prefix: URL prefix.
        """
        self.router.add_router(router, prefix=prefix)


class RootASGIApp:
    """Pykour application class."""

    SUPPORTED_METHODS = ["GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS", "HEAD"]

    def __init__(self):
        """Initialize Pykour application."""
        self.logger = logging.getLogger("pykour")

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = Request(scope, receive)
        response = Response(send)
        start_time = time.perf_counter()
        try:
            # Check if the scheme is supported
            if not self.is_supported_scheme(request):
                await self.handle_bad_request(request, response)
                return

            # Check if the method is supported
            if not self.is_supported_method(request):
                if self.logger.isEnabledFor(logging.DEBUG):
                    self.logger.debug(f"Unsupported HTTP Method: {request.method}")
                await self.handle_not_found(request, response)
                return

            # Check if the method is allowed
            if not self.is_method_allowed(request):
                if self.logger.isEnabledFor(logging.DEBUG):
                    self.logger.debug(f"Method not allowed: {request.method}")
                await self.handle_method_not_allowed(request, response)
                return

            # Process the request if the route exists
            if self.exists_route(request):
                self.append_path_params(request)
                await self.handle_request(self, request, response)
            else:
                if self.logger.isEnabledFor(logging.DEBUG):
                    self.logger.debug(f"Route not found: {request.path}")
                await self.handle_not_found(request, response)

        finally:
            end_time = time.perf_counter()
            write_access_log(request, response, (end_time - start_time) * 1000)

    @staticmethod
    def append_path_params(request: Request) -> None:
        """Append path parameters to the request."""
        app = request.app
        path = request.path
        method = request.method
        route = app.router.get_route(path, method)

        path_params = route.path_params
        request.path_params = path_params

    @staticmethod
    async def handle_bad_request(request: Request, response: Response) -> None:
        response.status = HTTPStatus.BAD_REQUEST
        response.content_type = "text/plain"
        response.content = HTTPStatus.BAD_REQUEST.phrase
        for accept in request.accept:
            if accept in ["application/json"]:
                response.content_type = accept
                response.content = json.dumps({"error": HTTPStatus.BAD_REQUEST.phrase})
                break
        await response.render()

    @staticmethod
    async def handle_not_found(request: Request, response: Response) -> None:
        response.status = HTTPStatus.NOT_FOUND
        response.content_type = "text/plain"
        response.content = HTTPStatus.NOT_FOUND.phrase
        for accept in request.accept:
            if accept in ["application/json"]:
                response.content_type = accept
                response.content = json.dumps({"error": HTTPStatus.NOT_FOUND.phrase})
                break
        await response.render()

    @staticmethod
    async def handle_method_not_allowed(request: Request, response: Response) -> None:
        response.status = HTTPStatus.METHOD_NOT_ALLOWED
        response.content_type = "text/plain"
        response.content = HTTPStatus.METHOD_NOT_ALLOWED.phrase
        for accept in request.accept:
            if accept in ["application/json"]:
                response.content_type = accept
                response.content = json.dumps({"error": HTTPStatus.METHOD_NOT_ALLOWED.phrase})
                break
        await response.render()

    @staticmethod
    async def handle_internal_server_error(request: Request, response: Response) -> None:
        response.status = HTTPStatus.INTERNAL_SERVER_ERROR
        response.content_type = "text/plain"
        response.content = HTTPStatus.INTERNAL_SERVER_ERROR.phrase
        for accept in request.accept:
            if accept in ["application/json"]:
                response.content_type = accept
                response.content = json.dumps({"error": HTTPStatus.INTERNAL_SERVER_ERROR.phrase})
                break
        await response.render()

    @staticmethod
    async def handle_http_exception(request: Request, response: Response, e: ex.HTTPException) -> None:
        response.status = e.status_code
        response.content_type = "text/plain"
        response.content = e.message
        for accept in request.accept:
            if accept in ["application/json"]:
                response.content_type = accept
                response.content = json.dumps({"error": e.message})
                break
        await response.render()

    @staticmethod
    async def handle_request(self, request: Request, response: Response):
        """Handle request for a route."""

        app = request.app
        route = app.router.get_route(request.path, request.method)
        route_fun, status_code = route.handler
        response.status = status_code

        # noinspection PyBroadException
        try:
            response_body = await call(route_fun, request, response)

            if response.status == HTTPStatus.NO_CONTENT:
                response.content = ""
                response.content_type = "text/plain"
            elif request.method == "OPTIONS":
                response.add_header("Allow", ", ".join(app.router.get_allowed_methods(request.path)))
                response.content = ""
            elif request.method == "HEAD":
                response.add_header("Content-Length", str(len(str(response_body))))
                response.content = ""
            elif isinstance(response_body, (dict, list)):
                response.content = json.dumps(response_body)
                response.content_type = "application/json"
            elif isinstance(response_body, str):
                response.content = response_body
                response.content_type = "text/plain"
            else:
                raise ValueError("Unsupported response type: %s" % type(response_body))

            await response.render()
        except ex.HTTPException as e:
            await self.handle_http_exception(request, response, e)
        except Exception as e:
            if self.logger.isEnabledFor(logging.DEBUG):
                self.logger.debug(f"Internal Server Error: {e}")
            await self.handle_internal_server_error(request, response)

    @staticmethod
    def is_supported_scheme(request: Request) -> bool:
        """Check if the scheme is supported."""
        return request.scheme in ["HTTP"]

    @staticmethod
    def is_supported_method(request: Request) -> bool:
        """Check if the method is supported."""
        return request.method in RootASGIApp.SUPPORTED_METHODS

    @staticmethod
    def is_method_allowed(request: Request) -> bool:
        """Check if the method is allowed for the given path."""
        app = request.app
        path = request.path
        method = request.method
        allowed_methods = app.router.get_allowed_methods(path)
        return allowed_methods == [] or method in allowed_methods

    @staticmethod
    def exists_route(request: Request) -> bool:
        """Check if the route exists."""
        app = request.app
        path = request.path
        method = request.method
        return app.router.exists(path, method)



from __future__ import annotations

import logging
import os
from typing import Any, Union, Optional, List

import yaml
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from pykour.logging import ACCESS_LEVEL_NO


def replace_placeholders(config: dict):
    for key, value in config.items():
        if isinstance(value, dict):
            replace_placeholders(value)
        elif isinstance(value, str):
            config[key] = os.path.expandvars(value)


class ConfigFileHandler(FileSystemEventHandler):
    def __init__(self, config: Config):
        self.config = config

    def on_modified(self, event):
        if event.src_path == self.config.filepath:
            self.config.load()
            print("Config file has been modified. Reloading...")


class Config:
    KEY_PYKOUR_LOGGING_LEVEL = "pykour.logging.level"
    KEY_PYKOUR_DATASOURCE_TYPE = "pykour.datasource.type"
    KEY_PYKOUR_DATASOURCE_URL = "pykour.datasource.url"
    KEY_PYKOUR_DATASOURCE_USERNAME = "pykour.datasource.username"
    KEY_PYKOUR_DATASOURCE_PASSWORD = "pykour.datasource.password"
    KEY_PYKOUR_DATASOURCE_POOL_MAX_CONNECTIONS = "pykour.datasource.pool.max-connections"

    def __init__(self, filepath=None):
        self.config = {}
        self._last_modified = 0.0
        if filepath is not None:
            self.filepath = os.path.abspath(filepath)
            self.load()
            self._setup_watchdog()

    def load(self):
        try:
            with open(self.filepath, "r") as file:
                print(f"Loading config file: {self.filepath}")
                content = yaml.safe_load(file)
                if isinstance(content, str):
                    print("Format of config file is invalid. Expected a yaml format.")
                    content = {}
                else:
                    replace_placeholders(content)
                self.config = content
            self._last_modified = os.path.getmtime(self.filepath)
        except FileNotFoundError:
            print(f"Config file not found: {self.filepath}")

    def reload(self):
        current_mtime = os.path.getmtime(self.filepath)
        if current_mtime > self._last_modified:
            print("Config file has been modified. Reloading...")
            self.load()

    def _setup_watchdog(self):
        event_handler = ConfigFileHandler(self)
        self.observer = Observer()
        self.observer.schedule(event_handler, path=os.path.dirname(self.filepath), recursive=False)
        self.observer.start()

    def get(self, key: str, default: Union[Any, None] = None):
        keys = key.split(".")
        d = self.config
        for k in keys:
            if k not in d:
                return default
            d = d[k]
        return d

    def get_int(self, key: str, default: Optional[int] = None) -> Optional[int]:
        value = self.get(key, default)
        if value is None:
            return default
        if isinstance(value, (int, float, str)):
            try:
                return int(value)
            except (ValueError, TypeError):
                raise ValueError(f"Cannot cast value of key '{key}' to int: {value}")
        else:
            raise ValueError(f"Cannot cast value of key '{key}' to int: {value}")

    def get_float(self, key: str, default: Optional[float] = None) -> Optional[float]:
        value = self.get(key, default)
        if value is None:
            return default
        if isinstance(value, (int, float, str)):
            try:
                return float(value)
            except (ValueError, TypeError):
                raise ValueError(f"Cannot cast value of key '{key}' to float: {value}")
        else:
            raise ValueError(f"Cannot cast value of key '{key}' to float: {value}")

    def get_log_levels(self) -> List[int]:
        log_levels = self.get(self.KEY_PYKOUR_LOGGING_LEVEL, "INFO, WARN, ERROR")
        level_names = [level.strip().upper() for level in log_levels.split(",")]

        level_numbers = []
        for level_name in level_names:
            try:
                level_number = getattr(logging, level_name)
                level_numbers.append(level_number)
            except AttributeError:
                raise ValueError(f"Unknown log level: {level_name}")

        level_numbers.append(ACCESS_LEVEL_NO)
        return level_numbers

    def get_datasource_type(self) -> str:
        return self.get(self.KEY_PYKOUR_DATASOURCE_TYPE, None)

    def get_datasource_url(self) -> str:
        return self.get(self.KEY_PYKOUR_DATASOURCE_URL, None)

    def get_datasource_username(self) -> str:
        return self.get(self.KEY_PYKOUR_DATASOURCE_USERNAME, None)

    def get_datasource_password(self) -> str:
        return self.get(self.KEY_PYKOUR_DATASOURCE_PASSWORD, None)

    def get_datasource_pool_max_connections(self) -> int:
        return self.get_int(self.KEY_PYKOUR_DATASOURCE_POOL_MAX_CONNECTIONS, 5)

    def __del__(self):
        if hasattr(self, "observer"):
            self.observer.stop()
            self.observer.join()

    def __str__(self):
        return yaml.dump(self.config, default_flow_style=False, allow_unicode=True, sort_keys=False)

    def __repr__(self):
        return self.__str__()



import logging
import re
import threading
from datetime import datetime
from http import HTTPStatus

from colorama import Fore, Style

from pykour.request import Request
from pykour.response import Response

from pykour.globals import thread_local

LOG_COLORS = {
    "INFO": Fore.GREEN,
    "WARN": Fore.YELLOW,
    "ERROR": Fore.RED,
    "TRACE": Fore.CYAN,
    "ACCESS": Fore.BLUE,
}

STATUS_COLORS = {
    "2xx": Fore.GREEN,
    "3xx": Fore.BLUE,
    "4xx": Fore.YELLOW,
    "5xx": Fore.RED,
}


class InterceptHandler(logging.Handler):
    def emit(self, record):
        pass


ACCESS_LEVEL_NO = 25
ACCESS_LEVEL_NAME = "ACCESS"


def access(self, message, *args, **kws):
    self._log(ACCESS_LEVEL_NO, message, args, **kws)


logging.addLevelName(ACCESS_LEVEL_NO, ACCESS_LEVEL_NAME)
logging.Logger.access = access  # type: ignore[attr-defined]


class CustomFormatter(logging.Formatter):
    converter = datetime.fromtimestamp  # type: ignore[assignment]

    def formatTime(self, record, datefmt=None):
        dt = self.converter(record.created)  # type: ignore[has-type]
        if datefmt:
            s = dt.strftime(datefmt)
        else:
            s = dt.isoformat(timespec="milliseconds")
        return s + self.format_time_zone()

    @staticmethod
    def format_time_zone():
        utc_offset = datetime.now().astimezone().strftime("%z")
        return utc_offset

    def format(self, record):
        record.levelname = f"{record.levelname:<6}"
        request_id = getattr(thread_local, "request_id", threading.get_ident())
        record.request_id = request_id
        return super().format(record)


class SpecificLevelsFilter(logging.Filter):
    def __init__(self, levels):
        super().__init__()
        self.levels = levels

    def filter(self, record):
        return record.levelno in self.levels


class CustomLogger(logging.Logger):
    def __init__(self, name):
        super().__init__(name)
        self.levels = [ACCESS_LEVEL_NO]

    def isEnabledFor(self, level):
        return level in self.levels


def setup_logging(log_levels=None) -> None:
    if log_levels is None:
        log_levels = [logging.INFO, logging.WARN, logging.ERROR, ACCESS_LEVEL_NO]

    # Suppress logging from Uvicorn and Gunicorn
    for _logger in ("uvicorn", "uvicorn.error", "uvicorn.access", "gunicorn.error"):
        logging_logger = logging.getLogger(_logger)
        logging_logger.handlers = [InterceptHandler()]
        logging_logger.propagate = False

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.NOTSET)
    level_color = LOG_COLORS.get(ACCESS_LEVEL_NAME, Fore.WHITE)
    formatter = CustomFormatter(
        f"{level_color}%(levelname)s{Style.RESET_ALL} [%(asctime)s] [%(request_id)s] %(message)s"
    )
    console_handler.setFormatter(formatter)
    levels_filter = SpecificLevelsFilter(levels=log_levels)
    console_handler.addFilter(levels_filter)

    logging.setLoggerClass(CustomLogger)

    logger = logging.getLogger("pykour")
    logger.setLevel(logging.NOTSET)
    logger.addHandler(console_handler)

    logger.levels = log_levels  # type: ignore[attr-defined]


def write_access_log(request: Request, response: Response, elapsed: float) -> None:
    """Write access log."""

    logger = logging.getLogger("pykour.access")

    category = f"{response.status // 100}xx"
    category_color = STATUS_COLORS.get(category, Fore.WHITE)

    client = request.client or "-"
    method = request.method or "-"
    path = request.path or "-"
    path = re.sub(r"/+", "/", path)
    scheme = request.scheme or "-"
    version = request.version or "-"
    status = response.status
    phrase = HTTPStatus(response.status).phrase
    content = response.content or ""

    logger.access(  # type: ignore[attr-defined]
        f"{client} - - {method} {path} {scheme}/{version} {category_color}{status} {phrase}{Style.RESET_ALL}"
        + f" {len(str(content))} {elapsed:.6f}",
    )



import gzip
import io
from typing import Callable, Union, List, Dict

from pykour.types import ASGIApp, Scope, Receive, Send
from pykour.middleware.base import BaseMiddleware


class GZipMiddleware(BaseMiddleware):
    """Middleware to GZIP responses."""

    def __init__(self, app: ASGIApp, minimum_size: int = 500):
        """
        Initialize the middleware with the application.

        Args:
            app: The ASGI application.
            minimum_size: The minimum size in bytes for responses to be gzipped.
        """
        super().__init__(app)
        self.minimum_size = minimum_size

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        """
        Call the middleware.

        Args:
            scope: The ASGI scope.
            receive: The ASGI receive channel.
            send: The ASGI send channel.
        """
        if scope["type"] == "http":
            headers = {k.lower(): v for k, v in scope["headers"]}
            if b"accept-encoding" in headers and b"gzip" in headers[b"accept-encoding"]:
                responder = GZipResponder(self.app, self.minimum_size)
                await responder(scope, receive, send)
            else:
                await self.app(scope, receive, send)
        else:
            await self.app(scope, receive, send)


class GZipResponder:
    def __init__(self, app: ASGIApp, minimum_size: int):
        self.app = app
        self.minimum_size = minimum_size
        self.body: List[bytes] = []
        self.send: Union[Callable, None] = None
        self.initial_message: Union[Dict, None] = None

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        self.send = send
        self.body = []
        self.initial_message = None

        async def send_wrapper(message):
            if message["type"] == "http.response.start":
                self.initial_message = message
            elif message["type"] == "http.response.body":
                self.body.append(message.get("body", b""))
                if not message.get("more_body", False):
                    body = b"".join(self.body)
                    if len(body) >= self.minimum_size:
                        gzip_buffer = io.BytesIO()
                        with gzip.GzipFile(fileobj=gzip_buffer, mode="wb") as f:
                            f.write(body)
                        body = gzip_buffer.getvalue()
                        self.initial_message["headers"].append((b"content-encoding", b"gzip"))
                    await self.send(self.initial_message)
                    await self.send({"type": "http.response.body", "body": body, "more_body": False})
            else:
                await self.send(message)

        await self.app(scope, receive, send_wrapper)


def gzip_middleware(minimum_size: int = 500):
    def middleware(app: ASGIApp):
        return GZipMiddleware(app, minimum_size=minimum_size)

    return middleware



from .base import BaseMiddleware
from .uuid import UUIDMiddleware, uuid_middleware
from .gzip import GZipMiddleware, gzip_middleware

__all__ = ["BaseMiddleware", "UUIDMiddleware", "uuid_middleware", "GZipMiddleware", "gzip_middleware"]



import inspect
import logging
from datetime import datetime
from typing import Any, Callable, Dict, get_origin
from enum import Enum

from pykour.config import Config
from pykour.db.connection import Connection
from pykour.request import Request
from pykour.response import Response
from pykour.schema import BaseSchema


def cast_to_type(value: Any, to_type: type) -> Any:
    if to_type == int:
        return int(value)
    elif to_type == float:
        return float(value)
    elif to_type == bool:
        return value.lower() in ["true", "1", "yes"]
    elif to_type == datetime:
        return datetime.strptime(value, "%Y-%m-%d")
    elif issubclass(to_type, Enum):
        try:
            return to_type[value]
        except KeyError:
            raise ValueError(f"{value} is not a valid {to_type.__name__}")
    else:
        return value


async def call(func: Callable, request: Request, response: Response) -> Any:
    sig = inspect.signature(func)
    logger = logging.getLogger("pykour")

    path_params = request.path_params
    app = request.scope.get("app")
    pool = app.pool
    conn = None

    bound_args: Dict[str, Any] = {}

    for param_name, param in sig.parameters.items():
        if isinstance(param.annotation, type) and issubclass(param.annotation, BaseSchema):
            bound_args[param_name] = param.annotation.from_dict(await request.json())
        elif param.annotation is dict or get_origin(param.annotation) is dict:
            bound_args[param_name] = await request.json()
        elif param.annotation is Request or param_name == "request" or param_name == "req":
            bound_args[param_name] = request
        elif param.annotation is Response or param_name == "response" or param_name == "res" or param_name == "resp":
            bound_args[param_name] = response
        elif param_name in path_params:
            bound_args[param_name] = cast_to_type(path_params[param_name], param.annotation)
        elif request.scope.get("app").config and (param.annotation is Config or param_name == "config"):
            bound_args[param_name] = request.get("app").config
        elif param.annotation is Connection or param_name == "conn" or param_name == "connection":
            if pool:
                if conn:
                    bound_args[param_name] = conn
                else:
                    conn = pool.get_connection()
                    bound_args[param_name] = conn
            else:
                bound_args[param_name] = None

    try:
        result = func(**bound_args)
        ret = None
        if inspect.iscoroutine(result):
            ret = await result
        else:
            ret = result

        if conn:
            conn.commit()
        return ret
    except Exception as e:
        if logger.isEnabledFor(logging.ERROR):
            logger.error(f"Error occurred while calling {func.__name__}: {e}")
        if conn:
            conn.rollback()
        raise e
    finally:
        if conn:
            pool.release_connection(conn)



from typing import Callable, Any


def validate(field_name: str) -> Callable:
    """Validator decorator.

    Args:
        field_name: The name of the field to validate.
    """

    def decorator(func: Callable[[Any], None]) -> Callable[[Any], None]:
        func.__validator__ = field_name  # type: ignore[attr-defined]
        return func

    return decorator



from __future__ import annotations

from http import HTTPStatus


class HTTPException(Exception):
    """HTTPException is a base class for HTTP exceptions."""

    def __init__(self, status_code: int, message: str | None = None) -> None:
        if message is None:
            message = HTTPStatus(status_code).phrase
        self.status_code = status_code
        self.message = message

    def __str__(self) -> str:
        return f"{self.status_code}: {self.message}"

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code} message={self.message})"


class ResourceNotFoundException(HTTPException):
    """ResourceNotFoundException is raised when a resource is not found."""

    def __init__(self, message: str | None = None) -> None:
        super().__init__(status_code=HTTPStatus.NOT_FOUND, message=message)



from http import HTTPStatus
from typing import Callable, Union, MutableMapping, Any, Awaitable


Scope = MutableMapping[str, Any]
Message = MutableMapping[str, Any]
Receive = Callable[[], Awaitable[Message]]
Send = Callable[[Message], Awaitable[None]]
ASGIApp = Callable[[Scope, Receive, Send], Awaitable[None]]
HTTPStatusCode = Union[int, HTTPStatus]



import logging
from uuid import uuid4

from pykour.middleware import BaseMiddleware
from pykour.types import Scope, Receive, Send, Message, ASGIApp

from pykour.globals import thread_local


class UUIDMiddleware(BaseMiddleware):

    def __init__(self, app, header_name="X-Request-ID"):
        super().__init__(app)
        self.header_name = header_name
        self.logger = logging.getLogger("uvicorn")

    async def process_request(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ):
        # generate default request_id
        request_id = str(uuid4())

        # check if X-Request-ID exists in headers
        for header in scope["headers"]:
            if header[0].decode("latin1") == self.header_name:
                request_id = header[1].decode("latin1")
                break

        # if not, add X-Request-ID to headers
        if not any(header[0].decode("latin1") == self.header_name for header in scope["headers"]):
            scope["headers"].append((self.header_name.encode("latin1"), request_id.encode("latin1")))

        scope["request_id"] = request_id
        thread_local.request_id = request_id
        if self.logger.isEnabledFor(logging.INFO):
            self.logger.info(f"{self.header_name}: {scope['request_id']}")

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        self.scope = scope
        self.send = send
        await self.process_request(scope, receive, send)
        await self.app(scope, receive, self.send_with_request_id)

    async def send_with_request_id(self, message: Message) -> None:
        if message["type"] == "http.response.start":
            message["headers"].append((self.header_name.encode("latin1"), self.scope["request_id"].encode()))
        await self.send(message)


def uuid_middleware(header_name="x-request-id"):
    def middleware(app: ASGIApp):
        return UUIDMiddleware(app, header_name=header_name)

    return middleware
